/* NOR flash interface
 * Copyright (C) 2014 Daniel Beer <dlbeer@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef NOR_H_
#define NOR_H_

#include <stdint.h>

/* nor_word_t: the smallest unit of data transfer during programming.
 * The programming interface will be fed a word at a time.
 */
typedef uint16_t nor_word_t;

#define NOR_WORD_BITS	16

/* nor_block_t: this is an opaque handle which is used to refer to a NOR
 * flash erase-unit.
 *
 * It might just be a pointer, or it could be a pointer to a structure
 * holding SPI bus information, for example.
 *
 * In this example, it's just a pointer to an array.
 */
typedef nor_word_t *nor_block_t;

/* Size of a NOR flash block, in words */
#define NOR_WORDS	256

/* Streaming read interface. For this interface, you need to define an
 * abstract type (nor_read_pointer_t) which tracks the state of an open
 * read stream.
 *
 * Reading is always done sequentially, with a begin and and end phase.
 *
 * In our example, reading is just array traversal.
 */
typedef int nor_read_pointer_t;

static inline void nor_read_begin(nor_block_t blk, nor_read_pointer_t *ptr)
{
	*ptr = 0;
}

static inline nor_word_t nor_read_word(nor_block_t blk,
				       nor_read_pointer_t *ptr)
{
	return blk[(*ptr)++];
}

static inline void nor_read_end(nor_block_t blk, nor_read_pointer_t *ptr)
{
	/* Nothing to do */
}

/* Streaming write interface. This is similar to the read interface, but
 * the begin phase should perform a block erase.
 *
 * In our example interface, we don't deal with real hardware -- we're
 * just writing to an array.
 */
typedef int nor_write_pointer_t;

static inline void nor_write_begin(nor_block_t blk, nor_write_pointer_t *ptr)
{
	*ptr = 0;
}

static inline void nor_write_word(nor_block_t blk, nor_write_pointer_t *ptr,
				  nor_word_t w)
{
	blk[(*ptr)++] = w;
}

static inline void nor_write_end(nor_block_t blk, nor_write_pointer_t *ptr)
{
	/* Nothing to do */
}

/* Shadow-page definitions. For a given block, this function should
 * return the block's shadow.
 *
 * In our example, the shadow block is the block immediately following.
 */
static inline nor_block_t nor_shadow(nor_block_t blk)
{
	return blk + NOR_WORDS;
}

#endif
