/* NOR shadow-paging
 * Copyright (C) 2014 Daniel Beer <dlbeer@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "shadow.h"
#include "nor.h"
#include "crc32.h"

#define SHADOW_MAGIC	((nor_word_t)0x34be5b72)
#define SHADOW_INVERSE	((nor_word_t)~SHADOW_MAGIC)

#define FILL_WORD	((nor_word_t)~0)

#define MSB(type, data) (((type)(data)) << ((sizeof(type) - 1) * 8))

static inline uint32_t crc_word(uint32_t crc, nor_word_t w)
{
	int i;

	for (i = 0; i < sizeof(nor_word_t); i++) {
		crc = crc32(crc, w);
		w >>= 8;
	}

	return crc;
}

static shadow_state_t scan_page(nor_block_t blk)
{
	nor_read_pointer_t ptr;
	nor_word_t magic;
	uint32_t crc = CRC32_INIT;
	size_t i;

	nor_read_begin(blk, &ptr);

	for (i = 0; i < SHADOW_PAYLOAD_WORDS; i++)
		crc = crc_word(crc, nor_read_word(blk, &ptr));

	magic = nor_read_word(blk, &ptr);
	crc = crc_word(crc, magic);

	for (i = 0; i < SHADOW_CHECKSUM_WORDS; i++)
		crc = crc_word(crc, nor_read_word(blk, &ptr));

	nor_read_end(blk, &ptr);

	if (crc)
		return 0;

	if (magic == SHADOW_MAGIC)
		return SHADOW_FLAG_VALID;
	if (magic == SHADOW_INVERSE)
		return SHADOW_FLAG_VALID | SHADOW_FLAG_INV_MAGIC;

	return 0;
}

shadow_state_t shadow_scan(nor_block_t blk)
{
	const shadow_state_t s0 = scan_page(blk);
	const shadow_state_t s1 = scan_page(nor_shadow(blk));
	const shadow_state_t valid = (s0 | s1) & SHADOW_FLAG_VALID;
	const shadow_state_t alt_block =
		(!(s0 & SHADOW_FLAG_VALID) ||
		 ((s0 & s1 & SHADOW_FLAG_VALID) &&
		  ((~(s0 ^ s1)) & SHADOW_FLAG_INV_MAGIC)))
		? SHADOW_FLAG_ALT_BLOCK : 0;
	const shadow_state_t inv_magic =
		(alt_block ? s1 : s0) & SHADOW_FLAG_INV_MAGIC;

	return valid | alt_block | inv_magic;
}

void shadow_read_begin(struct shadow_reader *r,
		       shadow_state_t st, nor_block_t blk)
{
	r->blk = (st & SHADOW_FLAG_ALT_BLOCK) ? nor_shadow(blk) : blk;
	r->word_buf = 0;
	r->ptr = 0;

	nor_read_begin(r->blk, &r->word_ptr);
}

size_t shadow_read(struct shadow_reader *r, void *buf_v, size_t len)
{
	uint8_t *buf = (uint8_t *)buf_v;
	size_t i;

	if (len > SHADOW_PAYLOAD_SIZE - r->ptr)
		len = SHADOW_PAYLOAD_SIZE - r->ptr;

	for (i = 0; i < len; i++) {
		if (!(r->ptr & (sizeof(nor_word_t) - 1)))
			r->word_buf = nor_read_word(r->blk, &r->word_ptr);

		buf[i] = r->word_buf;
		r->word_buf >>= 8;
		r->ptr++;
	}

	return len;
}

void shadow_write_begin(struct shadow_writer *w, shadow_state_t st,
			nor_block_t blk)
{
	w->blk = (st & SHADOW_FLAG_ALT_BLOCK) ? blk : nor_shadow(blk);
	w->word_buf = 0;
	w->ptr = 0;
	w->crc = CRC32_INIT;

	nor_write_begin(w->blk, &w->word_ptr);
}

static shadow_state_t state_after_write(shadow_state_t st)
{
	const shadow_state_t alt_block = (~st) & SHADOW_FLAG_ALT_BLOCK;
	const shadow_state_t inv_magic =
		(alt_block ? st : ~st) & SHADOW_FLAG_INV_MAGIC;

	return SHADOW_FLAG_VALID | alt_block | inv_magic;
}

shadow_state_t shadow_write_end(struct shadow_writer *w, shadow_state_t st)
{
	const shadow_state_t next_state = state_after_write(st);
	const nor_word_t magic = (next_state & SHADOW_FLAG_INV_MAGIC) ?
		SHADOW_INVERSE : SHADOW_MAGIC;

	/* Pad the last word */
	if (w->ptr & (sizeof(nor_word_t) - 1)) {
		while (w->ptr & (sizeof(nor_word_t) - 1)) {
			w->word_buf = (w->word_buf >> 8) |
				MSB(nor_word_t, FILL_WORD);
			w->ptr++;
		}

		nor_write_word(w->blk, &w->word_ptr, w->word_buf);
		w->crc = crc_word(w->crc, w->word_buf);
	}

	/* Fill the remainder */
	while (w->ptr < SHADOW_PAYLOAD_SIZE) {
		nor_write_word(w->blk, &w->word_ptr, FILL_WORD);
		w->crc = crc_word(w->crc, FILL_WORD);
		w->ptr += sizeof(nor_word_t);
	}

	/* Write the magic word */
	nor_write_word(w->blk, &w->word_ptr, magic);
	w->crc = crc_word(w->crc, magic);
	w->ptr += sizeof(nor_word_t);

	/* Write the checksum */
	while (w->ptr < (NOR_WORDS * sizeof(nor_word_t))) {
		nor_write_word(w->blk, &w->word_ptr, w->crc);
#if NOR_WORD_BITS < 32
		w->crc >>= NOR_WORD_BITS;
#endif
		w->ptr += sizeof(nor_word_t);
	}

	nor_write_end(w->blk, &w->word_ptr);

	/* Verify the write */
	if (scan_page(w->blk) == (next_state & ~SHADOW_FLAG_ALT_BLOCK))
		return next_state;

	return st;
}

size_t shadow_write(struct shadow_writer *w, const void *buf_v, size_t len)
{
	const uint8_t *buf = (const uint8_t *)buf_v;
	size_t i;

	if (len > SHADOW_PAYLOAD_SIZE - w->ptr)
		len = SHADOW_PAYLOAD_SIZE - w->ptr;

	for (i = 0; i < len; i++) {
		w->word_buf >>= 8;
		w->word_buf |= MSB(nor_word_t, buf[i]);
		w->ptr++;

		if (!(w->ptr & (sizeof(nor_word_t) - 1))) {
			nor_write_word(w->blk, &w->word_ptr, w->word_buf);
			w->crc = crc_word(w->crc, w->word_buf);
		}
	}

	return len;
}
