/* NOR shadow-paging
 * Copyright (C) 2014 Daniel Beer <dlbeer@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SHADOW_H_
#define SHADOW_H_

#include <stdint.h>
#include <stddef.h>
#include "nor.h"

/* Words occupied by checksum */
#define SHADOW_CHECKSUM_WORDS	((sizeof(nor_word_t) + 3) / sizeof(nor_word_t))
#define SHADOW_HEADER_WORDS	(SHADOW_CHECKSUM_WORDS + 1)

/* Payload words and size */
#define SHADOW_PAYLOAD_WORDS	(NOR_WORDS - SHADOW_HEADER_WORDS)
#define SHADOW_PAYLOAD_SIZE	(SHADOW_PAYLOAD_WORDS * sizeof(nor_word_t))

/* Shadow-page state: this tells us which of the pair of pages is the
 * most recent version, and which magic number phase we're currently on.
 */
#define SHADOW_FLAG_VALID	0x01
#define SHADOW_FLAG_ALT_BLOCK	0x02
#define SHADOW_FLAG_INV_MAGIC	0x04

typedef uint8_t shadow_state_t;

/* Given the representative block of a shadow-page pair, scan the pair
 * and determine the state.
 */
shadow_state_t shadow_scan(nor_block_t blk);

/* Do we have valid data to begin with? */
static inline int shadow_has_data(shadow_state_t s)
{
	return s & SHADOW_FLAG_VALID;
}

/* Streaming read from the current page */
struct shadow_reader {
	nor_block_t		blk;
	nor_read_pointer_t	word_ptr;
	nor_word_t		word_buf;
	size_t			ptr;
};

void shadow_read_begin(struct shadow_reader *r,
		       shadow_state_t st, nor_block_t blk);

static inline void shadow_read_end(struct shadow_reader *r)
{
	nor_read_end(r->blk, &r->word_ptr);
}

static inline size_t shadow_read_ptr(const struct shadow_reader *r)
{
	return r->ptr;
}

size_t shadow_read(struct shadow_reader *r, void *buf, size_t);

/* Streaming write to the next page version */
struct shadow_writer {
	nor_block_t		blk;
	nor_write_pointer_t	word_ptr;
	nor_word_t		word_buf;
	size_t			ptr;
	uint32_t		crc;
};

void shadow_write_begin(struct shadow_writer *w, shadow_state_t st,
			nor_block_t blk);
shadow_state_t shadow_write_end(struct shadow_writer *w, shadow_state_t st);

static inline size_t shadow_write_ptr(const struct shadow_writer *w)
{
	return w->ptr;
}

size_t shadow_write(struct shadow_writer *w, const void *buf, size_t len);

#endif
